/* Extended precision math routines                                          */

/*--- Microsoft C 5.0 compiler include files                                 */

#include <math.h>
#include <memory.h>
#include <stdlib.h>

/*--- other includes                                                         */

#include "driver.h"

#define EXTMATH
#include "extmath.h"

/*---------------------------------------------------------------------------*/
/* Extended precision equivalents for decimal fractions                      */
/* Each entry is the extended precision equivalent of the corresponding      */
/*  decimal fraction 1.0, 0.1, 0.01, 0.001, and so forth                     */
/* These are stored high order byte first and flipped in SetupMath           */
/*  ... simply because flipping them during transcription is hard to do      */

unsigned char binfracts[DBLPREC][NUMLEN] = {
  { '\x10','\x00','\x00','\x00','\x00','\x00','\x00','\x00'}, /*  0 1.0      */
  { '\x01','\x99','\x99','\x99','\x99','\x99','\x99','\x99'}, /*  1 0.1      */
  { '\x00','\x28','\xf5','\xc2','\x8f','\x5c','\x28','\xf5'}, /*  2 0.01     */
  { '\x00','\x04','\x18','\x93','\x74','\xbc','\x6a','\x7e'}, /*  3 0.001    */
  { '\x00','\x00','\x68','\xdb','\x8b','\xac','\x71','\x0c'}, /*  4 etc      */
  { '\x00','\x00','\x0a','\x7c','\x5a','\xc4','\x71','\xb4'}, /*  5          */
  { '\x00','\x00','\x01','\x0c','\x6f','\x7a','\x0b','\x5e'}, /*  6          */
  { '\x00','\x00','\x00','\x1a','\xd7','\xf2','\x9a','\xbc'}, /*  7          */
  { '\x00','\x00','\x00','\x02','\xaf','\x31','\xdc','\x46'}, /*  8          */
  { '\x00','\x00','\x00','\x00','\x44','\xb8','\x2f','\xa0'}, /*  9          */
  { '\x00','\x00','\x00','\x00','\x06','\xdf','\x37','\xf6'}, /* 10          */
  { '\x00','\x00','\x00','\x00','\x00','\xaf','\xeb','\xff'}, /* 11          */
  { '\x00','\x00','\x00','\x00','\x00','\x11','\x97','\x99'}, /* 12          */
  { '\x00','\x00','\x00','\x00','\x00','\x01','\xc2','\x5c'}, /* 13          */
  { '\x00','\x00','\x00','\x00','\x00','\x00','\x2d','\x09'}, /* 14          */
  { '\x00','\x00','\x00','\x00','\x00','\x00','\x04','\x80'}  /* 15          */
};

/*---------------------------------------------------------------------------*/
/* Test driver                                                               */

#if 0

EXTNUM x;
EXTNUM y;
EXTNUM z;

unsigned char *result;
DOUBLEBITS d = {2.0};
double xd,yd,zd;

int main(void) {

  SetupMath();

  while (0) {

     d.dbl = 8.2240699441899e-8;

     CvtDbl2Ext(y,d.dbl);
     CvtDbl2FullExt(z,d.dbl);

     yd = CvtExt2Dbl(y);
     zd = CvtExt2Dbl(z);

  }

  while (1) {

     CvtDbl2FullExt(x,xd);
     CvtDbl2FullExt(y,yd);

     memcpy(z,MultiplyExt(x,y),NUMLEN);

     zd = CvtExt2Dbl(z);

  }


  return 0;

}
#endif


/*---------------------------------------------------------------------------*/
/* Set up the conversion table                                               */
/* The table starts out in readable order, which we flip just once           */


void SetupMath(void) {

  int ctr;

  for (ctr = 0; ctr < DBLPREC; ctr++) {
     SwapExt(binfracts[ctr]);
  }

}


/*---------------------------------------------------------------------------*/
/* Add an extended precision number to the extended accumulator              */
/* Incoming pointer is to the low-order byte at the lowest address           */
/* Returns pointer to the result in the extended accumulator                 */
/* We ignore the ugly issue of overflows and suchlike                        */

unsigned char *AddExt(unsigned char *addend) {

  int index;
  unsigned int accum;

  accum = 0;

  for (index = 0; index < NUMLEN-1; index++) {   /* do all but last byte     */
     accum += ((unsigned int) ExtAccum[index]) + ((unsigned int) *addend);
     ExtAccum[index] = (unsigned char) accum;
     accum >>= 8;
     addend++;
  }

  accum += ((unsigned int) ExtAccum[index]) + ((unsigned int) *addend);
  ExtAccum[index] = (unsigned char) accum;

  return ExtAccum;

} /* AddExt */



/*---------------------------------------------------------------------------*/
/* Find 2's complement of extended precision number                          */
/* Puts result in extargneg                                                  */
/* Ignores overflow and other oddities                                       */

void NegExt(unsigned char *extargneg,unsigned char *extargin) {

  int ctr;
  unsigned int accum;

  accum = 1;
  for (ctr = 1; ctr <= NUMLEN; ctr++) {
     accum += 0x00ff ^ (unsigned int) *extargin;
     *extargneg = (unsigned char) accum;
     accum >>= 8;
     extargin++;
     extargneg++;
  }

  return;

} /* NegExt */

/*---------------------------------------------------------------------------*/
/* Find absolute value of extended precision number                          */
/* Puts result in extargabs                                                  */
/* Checks the sign bit and flips the result if needed                        */
/* Ignores overflow and other oddities                                       */

void AbsExt(unsigned char *extargabs,unsigned char *extargin) {

  if (*(extargin+NUMLEN-1) & 0x80)
     NegExt(extargabs,extargin);
  else
     memcpy(extargabs,extargin,NUMLEN);

  return;

} /* AbsExt */


/*---------------------------------------------------------------------------*/
/* Multiply two extended precision numbers                                   */
/* Result ends up in the high end of the extended accumulator                */
/* This is rather poky, but good enough for occasional use...                */
/* We blithely ignore issues of overflow and suchlike                        */

unsigned char *MultiplyExt(unsigned char *mcandin, unsigned char *mpyin) {

  int mpyndx;
  int mcandndx;
  int cyndx;

  unsigned int accum;
  unsigned int saccum;

  EXTNUM mcand;
  EXTNUM mpy;

/*--- set up the arguments and product buffer                                */

  AbsExt(mcand,mcandin);
  AbsExt(mpy,mpyin);                   /* this is OK if squaring             */
  memset(ExtAccum,'\0',2*NUMLEN);

/*--- run the multiply!                                                      */
/*    the confusion comes from propagating the carries                       */

  for (mpyndx = 0; mpyndx < NUMLEN; mpyndx++) {
     accum = 0;
     for (mcandndx = 0; mcandndx < NUMLEN; mcandndx++) {
        accum += (unsigned int) mcand[mcandndx] * (unsigned int) mpy[mpyndx];
        accum += (unsigned int) ExtAccum[mpyndx+mcandndx];
        ExtAccum[mpyndx+mcandndx] = (unsigned char) accum;
        for (cyndx = mpyndx+mcandndx+1; cyndx < (2*NUMLEN); cyndx++) {
           accum >>= 8;
           accum += ExtAccum[cyndx];
           ExtAccum[cyndx] = (unsigned char) accum;
        }
     }
  }

/*--- round the high part to NUMLEN bits                                     */
/*    there's an extra 4 bits on the high end, so the fix is four bits right */

  if (ExtAccum[NUMLEN-1] & 0x08) {
     accum = 0x10;
     for (cyndx = NUMLEN-1; cyndx < (2*NUMLEN) ; cyndx++) {
        accum += ExtAccum[cyndx];
        ExtAccum[cyndx] = (unsigned char) accum;
        accum >>= 8;
     }
  }


/*--- align the high part by shifting left four bits                         */

  saccum = (unsigned int) ExtAccum[NUMLEN-1];
  saccum >>= 4;

  for (cyndx = NUMLEN; cyndx < (2*NUMLEN); cyndx++) {
     accum = (unsigned int) ExtAccum[cyndx];
     accum = (accum << 4) | saccum;
     ExtAccum[cyndx] = (unsigned char) accum;
     saccum = accum >> 8;
  }

/*--- make high part negative if incoming signs differed                     */

  if ((*(mcandin+NUMLEN-1) & 0x80) != (*(mpyin+NUMLEN-1) & 0x80)) {
     memcpy(mcand,ExtAccum+NUMLEN,NUMLEN);
     NegExt(ExtAccum+NUMLEN,mcand);
  }

  return ExtAccum+NUMLEN;

} /* MultiplyExt */


/*---------------------------------------------------------------------------*/
/* Convert a double-precision value to extended fixed point                  */
/* Catches overflows into sign bit, sticks at max value                      */
/* Converts at maximum precision in the extended buffer, using decimal       */
/*  equivalent of each digit in double value...  we hope                     */

void CvtDbl2FullExt(unsigned char *extval, double dblval) {

  int digit;
  int reps;
  int dblsign;
  DOUBLEBITS d;
  double dblint,dblfract;

  d.dbl = dblval;                      /* set up union with double value     */

  dblsign = d.bits.signbit;            /* force double to positive           */
  d.bits.signbit = 0;

  if (d.dbl >= 8.0) {                  /* catch overflow                     */
     memset(ExtAccum,'\xff',NUMLEN-1);
     ExtAccum[NUMLEN-1] = 0x3f;                  /* force to max value       */
  }
  else {
     memset(ExtAccum,'\0',NUMLEN);               /* zap accumulator          */
     for (digit = 0; digit < DBLPREC; digit++) {
        dblfract = modf(d.dbl,&dblint);
        for (reps = 0; reps < (int) dblint; reps++) {
           AddExt(binfracts[digit]);
        }
        d.dbl = 10.0 * dblfract;                 /* extract next digit       */
     }

  }

  if (dblsign) {
     NegExt(ExtAccum+NUMLEN,ExtAccum);           /* complement & set return  */
     memcpy(extval,ExtAccum+NUMLEN,NUMLEN);
  }
  else {
     memcpy(extval,ExtAccum,NUMLEN);             /* just set return value    */
  }

  return;

} /* CvtDbl2FullExt */



/*---------------------------------------------------------------------------*/
/* Convert an extended fixed point value to double floating                  */
/* Returns the double value itself                                           */
/* Some bits will get dropped in the process                                 */

double CvtExt2Dbl(EXTNUM extval) {

  int index;
  int expshift;
  unsigned int accum;
  unsigned int saccum;
  DOUBLEBITS d;

  memset(&d.dbl,'\0',NUMLEN);          /* clear the decks                    */

  if (extval[NUMLEN-1] & 0x80) {       /* get positive in accumulator        */
     d.bits.signbit = 1;
     NegExt(ExtAccum,extval);
  }
  else {
     memcpy(ExtAccum,extval,NUMLEN);
  }

  accum = 0;                           /* if exactly zero, we're done!       */
  for (index = 0; index < NUMLEN; index++) {
     accum |= ExtAccum[index];
  }


  if (accum != 0) {
     expshift = 0;                     /* normalize first 1 bit location     */

     while (ExtAccum[NUMLEN-1] & 0xE0) {         /* right shift normalize    */
        saccum = 0;
        for (index = NUMLEN-1; index > -1; index--) {
           accum = (unsigned int) ExtAccum[index];
           accum = _rotr((accum | saccum),1);
           ExtAccum[index] = (unsigned char) accum;
           saccum = (0x8000 & accum) >> 7;
        }
        expshift++;
     }

     while (!(ExtAccum[NUMLEN-1] & 0x10)) {      /* left shift normalize     */
        saccum = 0;
        for (index = 0;index < NUMLEN; index++) {
           accum = (unsigned int) ExtAccum[index];
           accum = (accum << 1) | saccum;
           ExtAccum[index] = (unsigned char) accum;
           saccum = accum >> 8;
        }
        expshift--;
     }

     d.bits.exponent = expshift + 1023;          /* get excess 1023 exponent */

     d.bits.highmant = 0x000f & (unsigned int) ExtAccum[NUMLEN-1];

     memcpy(d.bits.mantissa,ExtAccum+1,6);       /* transfer useful bits     */

  }

  return d.dbl;

} /* CvtExt2Dbl */



/*---------------------------------------------------------------------------*/
/* Convert a double-precision value to extended fixed point                  */
/* Catches overflows into sign bit, sticks at max value                      */
/* The fact that double values have only 53 bits of precision forces low 0's */

void CvtDbl2Ext(unsigned char *extval, double dblval) {

  int index;
  int expshift;
  unsigned int accum;
  unsigned int saccum;
  DOUBLEBITS d;

  d.dbl = dblval;                      /* put it into the union              */

  expshift = d.bits.exponent - 1023;   /* convert exponent to shift amount   */

  if (expshift > 2) {                  /* catch overflow                     */
     memset(ExtAccum,'\xff',NUMLEN-1);
     ExtAccum[NUMLEN-1] = 0x3f;                  /* force to max value       */
     expshift = 0;                               /* and do no shifting       */
  }
  else {
     ExtAccum[0] = 0;                            /* fill unused bits         */
     memcpy(ExtAccum+1,d.bits.mantissa,6);       /* transfer low mantissa    */
     ExtAccum[NUMLEN-1] = (unsigned char) d.bits.highmant; /* high mantissa  */
     ExtAccum[NUMLEN-1] |= 0x10;                 /* add implied "1" bit      */
  }

  while (expshift > 0) {                         /* handle left shifts       */
     saccum = 0;
     for (index = 0;index < NUMLEN; index++) {
        accum = (unsigned int) ExtAccum[index];
        accum = (accum << 1) | saccum;
        ExtAccum[index] = (unsigned char) accum;
        saccum = accum >> 8;
     }
     expshift--;
  }

  while (expshift < 0) {                         /* handle right shifts      */
     saccum = 0;
     for (index = NUMLEN-1; index > -1; index--) {
        accum = (unsigned int) ExtAccum[index];
        accum = _rotr((accum | saccum),1);
        ExtAccum[index] = (unsigned char) accum;
        saccum = (0x8000 & accum) >> 7;
     }
     expshift++;
  }

  if (d.bits.signbit) {
     NegExt(ExtAccum+NUMLEN,ExtAccum);           /* complement & set return  */
     memcpy(extval,ExtAccum+NUMLEN,NUMLEN);
  }
  else {
     memcpy(extval,ExtAccum,NUMLEN);             /* just set return value    */
  }

  return;

} /* CvtDbl2Ext */



/*---------------------------------------------------------------------------*/
/* Multiply an extended fixed point value by 2.0                             */
/* This uses a simple shift and ignores overflow and suchlike                */

void Mpy2Ext(unsigned char *extval) {

  int ctr;
  unsigned int temp;

  temp = 0;                                      /* set bottom bit           */

  for (ctr=1; ctr<=NUMLEN; ctr++) {
     temp |= (unsigned int) *extval;             /* combine byte with bit    */
     temp = _rotl(temp,1);                       /* align bit in low slot    */
     *extval = (unsigned char) temp;             /* set low byte             */
     temp = (temp << 7) & 0x8000;                /* realign bit              */
     extval++;                                   /* tick to next byte        */
  }

}

/*---------------------------------------------------------------------------*/
/* Invert the byte order of an extended fixed point item...                  */
/* The Mandelbrot Engine stores these things high byte first                 */

void SwapExt(unsigned char *extval) {

  unsigned char *p1;
  unsigned char temp;
  int ctr;

  p1 = extval + NUMLEN - 1;

  for (ctr = 1; ctr <= (NUMLEN/2); ctr++) {
     temp = *extval;
     *extval = *p1;
     *p1 = temp;
     extval++;
     p1--;
  }

}
